/*
 * indicator-network - user interface for connman
 * Copyright 2010 Canonical Ltd.
 *
 * Authors:
 * Kalle Valo <kalle.valo@canonical.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "connman-manager.h"

#include <string.h>

#include "connman-manager-xml.h"
#include "connman.h"
#include "marshal.h"
#include "connman-service.h"

G_DEFINE_TYPE(ConnmanManager, connman_manager, G_TYPE_OBJECT)

#define GET_PRIVATE(o) \
  (G_TYPE_INSTANCE_GET_PRIVATE((o), CONNMAN_TYPE_MANAGER, \
			       ConnmanManagerPrivate))

typedef struct _ConnmanManagerPrivate ConnmanManagerPrivate;

struct _ConnmanManagerPrivate {
  /*
   * I'm not sure if a hashtable is the best choise, especially because we
   * need traverse it quite often. But currently the number of services is
   * so low that it doesn't really matter.
   */
  GHashTable *services;

  GDBusProxy *proxy;
  gboolean connected;
  gboolean initialised;

  guint watch_id;

  /* used only by get_services() */
  ConnmanService **array;

  /* current default service, visible to clients */
  ConnmanService *default_service;

  /*
   * the new default service but waiting for it to be ready, not visible to
   * clients
   */
  ConnmanService *pending_default_service;
};

enum {
  SERVICE_ADDED_SIGNAL,
  SERVICE_REMOVED_SIGNAL,
  LAST_SIGNAL,
};

static guint signals[LAST_SIGNAL] = { 0 };

enum
{
  /* reserved */
  PROP_0,

  PROP_CONNECTED,
  PROP_DEFAULT_SERVICE,
};

static void connect_service_cb(GObject *object, GAsyncResult *res,
			       gpointer user_data)
{
  GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT(user_data);
  GDBusProxy *proxy = G_DBUS_PROXY(object);
  GError *error = NULL;

  g_dbus_proxy_call_finish(proxy, res, &error);

  if (error != NULL) {
    g_simple_async_result_set_from_error(simple, error);
    g_error_free(error);
    goto out;
  }

 out:
  g_simple_async_result_complete(simple);
  g_object_unref(simple);
}

void connman_manager_connect_service(ConnmanManager *self,
				     ConnmanServiceType type,
				     ConnmanServiceMode mode,
				     ConnmanServiceSecurity security,
				     guint8 *ssid,
				     guint ssid_len,
				     GCancellable *cancellable,
				     GAsyncReadyCallback callback,
				     gpointer user_data)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  gchar ssid_buf[ssid_len + 1];
  GSimpleAsyncResult *simple;
  GVariantBuilder *builder;
  GVariant *parameters;

  g_return_if_fail(CONNMAN_IS_MANAGER(self));
  g_return_if_fail(priv != NULL);

  simple = g_simple_async_result_new(G_OBJECT(self), callback,
                                      user_data,
				     connman_manager_connect_service);

  /*
   * FIXME: ssid can contain any character possible, including null, but
   * for now null terminate the buffer with null and send it as a string to
   * connman. But this needs to be properly fixed at some point.
   */
  memcpy(ssid_buf, ssid, ssid_len);
  ssid_buf[ssid_len] = '\0';

  builder = g_variant_builder_new(G_VARIANT_TYPE("a{sv}"));
  g_variant_builder_add(builder, "{sv}", "Type",
			g_variant_new_string(connman_service_type2str(type)));
  g_variant_builder_add(builder, "{sv}", "Mode",
			g_variant_new_string(connman_service_mode2str(mode)));
  g_variant_builder_add(builder, "{sv}", "Security",
			g_variant_new_string(connman_service_security2str(security)));
  g_variant_builder_add(builder, "{sv}", "SSID",
			g_variant_new_string(ssid_buf));

  parameters = g_variant_new("(a{sv})", builder);

  /* FIXME: cancel the call on dispose */
  g_dbus_proxy_call(priv->proxy, "ConnectService", parameters,
  		    G_DBUS_CALL_FLAGS_NONE, CONNECT_TIMEOUT, cancellable,
  		    connect_service_cb, simple);

  /* FIXME: doesn't builder leak here? */
}

void connman_manager_connect_service_finish(ConnmanManager *self,
					    GAsyncResult *res,
					    GError **error)
{
  GSimpleAsyncResult *simple = G_SIMPLE_ASYNC_RESULT(res);

  g_return_if_fail(CONNMAN_IS_MANAGER(self));
  g_return_if_fail(g_simple_async_result_get_source_tag(simple) ==
		   connman_manager_connect_service);

  g_simple_async_result_propagate_error(simple, error);
}

gboolean connman_manager_get_connected(ConnmanManager *self)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  g_return_val_if_fail(CONNMAN_IS_MANAGER(self), FALSE);
  g_return_val_if_fail(priv != NULL, FALSE);

  return priv->connected;
}

static void iterate_create_list(gpointer key, gpointer value,
				gpointer user_data)
{
  GList **l = user_data;
  ConnmanService *service = value;

  *l = g_list_append(*l, service);
}

/**
 * connman_manager_get_services:
 *
 * @self: self
 *
 * Returns: (array zero-terminated=1) (array length=false) (element-type Connman.Service) (transfer none): Null terminated list of services,
 * owned by ConnmanManager. Caller must create a reference if it uses them.
 */
ConnmanService **connman_manager_get_services(ConnmanManager *self)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  ConnmanService *service;
  GList *l;
  gint i;

  g_return_val_if_fail(CONNMAN_IS_MANAGER(self), NULL);
  g_return_val_if_fail(priv != NULL, NULL);

  if (!priv->connected)
    return NULL;

  if (priv->services == NULL)
    return NULL;

  l = NULL;

  g_hash_table_foreach(priv->services, iterate_create_list, &l);

  g_free(priv->array);

  /* allocate space for the pointers and the null termination */
  priv->array = g_malloc((sizeof(ConnmanService *) * g_list_length(l)) + 1);

  if (priv->array == NULL)
    return NULL;

  for (i = 0; l != NULL; l = l->next) {
    service = l->data;

    /*
     * return only services which are ready yet, there will be
     * a separate service added signal for the ones which are not ready yet
     */
    if (!connman_service_is_ready(service))
      continue;

    priv->array[i++] = service;
  }

  priv->array[i] = NULL;

  g_list_free(l);
  l = NULL;

  return priv->array;
}

/**
 * connman_manager_get_service:
 *
 * @self: self
 * @path: path
 *
 * Returns: (transfer none): The service or null if not found.
 */
ConnmanService *connman_manager_get_service(ConnmanManager *self,
					    const gchar *path)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  g_return_val_if_fail(CONNMAN_IS_MANAGER(self), NULL);
  g_return_val_if_fail(priv != NULL, NULL);

  return g_hash_table_lookup(priv->services, path);
}

/**
 * connman_manager_get_default_service:
 *
 * @self: self
 *
 * Returns: (transfer none): The service or null if not found.
 */
ConnmanService *connman_manager_get_default_service(ConnmanManager *self)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  g_return_val_if_fail(CONNMAN_IS_MANAGER(self), NULL);
  g_return_val_if_fail(priv != NULL, NULL);

  return priv->default_service;
}

static void update_default_service(ConnmanManager *self)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  gboolean changed = FALSE;
  ConnmanService *pending;

  g_return_if_fail(priv != NULL);

  pending = priv->pending_default_service;

  if (pending == NULL) {
    if (priv->default_service == NULL)
      goto out;

    /* remove current default service because there aren't any services */
    g_object_unref(priv->default_service);
    priv->default_service = NULL;
    changed = TRUE;
    goto check;
  }

  /* no need to do anything if the service is already the default */
  if (priv->default_service == pending)
    goto check;

  if (!connman_service_is_ready(pending))
    /* wait for the ready signal from the service */
    goto check;

  switch(connman_service_get_state(pending)) {
  case CONNMAN_SERVICE_STATE_IDLE:
  case CONNMAN_SERVICE_STATE_FAILURE:
  case CONNMAN_SERVICE_STATE_ASSOCIATION:
  case CONNMAN_SERVICE_STATE_CONFIGURATION:
  case CONNMAN_SERVICE_STATE_DISCONNECT:
    /* pending isn't connected and we can't use it as default, yet  */
    goto check;
  case CONNMAN_SERVICE_STATE_READY:
  case CONNMAN_SERVICE_STATE_LOGIN:
  case CONNMAN_SERVICE_STATE_ONLINE:
    break;
  }

  /* pending service is now ready, let's switch to it */
  if (priv->default_service != NULL)
    g_object_unref(priv->default_service);

  /* the ownership moves from pending_default_service to default_service */
  priv->default_service = pending;
  priv->pending_default_service = NULL;
  changed = TRUE;

 check:
  if (priv->default_service == NULL)
    goto out;

  /*
   * Let's check the state of the default service and whether it needs to
   * be removed.
   */
  switch(connman_service_get_state(priv->default_service)) {
  case CONNMAN_SERVICE_STATE_IDLE:
  case CONNMAN_SERVICE_STATE_FAILURE:
  case CONNMAN_SERVICE_STATE_ASSOCIATION:
  case CONNMAN_SERVICE_STATE_CONFIGURATION:
  case CONNMAN_SERVICE_STATE_DISCONNECT:
    /* remove the default service because it's not valid anymore */
    g_object_unref(priv->default_service);
    priv->default_service = NULL;
    changed = TRUE;
    break;
  case CONNMAN_SERVICE_STATE_READY:
  case CONNMAN_SERVICE_STATE_LOGIN:
  case CONNMAN_SERVICE_STATE_ONLINE:
    break;
  }

 out:
  /* don't send any signals until connected to connmand */
  if (priv->connected && changed)
    g_object_notify(G_OBJECT(self), "default-service");
}

static void iterate_service_ready(gpointer key, gpointer value,
				  gpointer user_data)
{
  gboolean *not_ready = user_data;
  ConnmanService *service = value;

  if (!connman_service_is_ready(service))
    *not_ready = TRUE;
}

static gboolean all_services_ready(ConnmanManager *self)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  gboolean not_ready = FALSE;

  g_hash_table_foreach(priv->services, iterate_service_ready, &not_ready);

  return !not_ready;
}

static void service_ready(ConnmanService *service, GParamSpec *pspec,
			  gpointer user_data)
{
  ConnmanManager *self = CONNMAN_MANAGER(user_data);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  if (!connman_service_is_ready(service))
    return;

  if (priv->pending_default_service != NULL &&
      priv->pending_default_service == service)
    update_default_service(self);

  /*
   * Send service added signals only after we are connected. The client
   * needs to call get_services() when it has received the connected
   * signal.
   */
  if (!priv->connected) {
    if (all_services_ready(self)) {
      priv->connected = TRUE;
      g_object_notify(G_OBJECT(self), "connected");
    }

    return;
  }

  g_signal_emit(self, signals[SERVICE_ADDED_SIGNAL], 0, service);
}

static void service_state_notify(ConnmanService *service, GParamSpec *pspec,
				 gpointer user_data)
{
  ConnmanManager *self = CONNMAN_MANAGER(user_data);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  if (priv->pending_default_service == service ||
      priv->default_service == service)
    update_default_service(self);
}

static ConnmanService *create_service(ConnmanManager *self,
				      const gchar *path)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  ConnmanService *service;

  service = connman_service_new(path);

  g_hash_table_insert(priv->services, g_strdup(path), service);

  /* FIXME: disconnect ready handler */
  g_signal_connect(service, "notify::ready", G_CALLBACK(service_ready),
		   self);
  g_signal_connect(service, "notify::state",
		   G_CALLBACK(service_state_notify), self);

  return service;
}

static void remove_service(ConnmanManager *self, ConnmanService *service)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  const gchar *path = connman_service_get_path(service);

  /*
   * Send service removed signals only when connected, connected
   * property with false value implies that services are lost.
   */
  if (priv->connected)
    g_signal_emit(self, signals[SERVICE_REMOVED_SIGNAL], 0, path);

  g_signal_handlers_disconnect_by_func(service,
				       G_CALLBACK(service_state_notify),
				       self);

  g_object_unref(service);
}

static GHashTable *create_services_list(ConnmanManager *self)
{
  return g_hash_table_new_full(g_str_hash, g_str_equal, g_free, NULL);
}

static void notify_service_removal(gpointer key, gpointer value,
				   gpointer user_data)
{
  ConnmanManager *self = CONNMAN_MANAGER(user_data);
  ConnmanService *service = CONNMAN_SERVICE(value);

  remove_service(self, service);
}

static void remove_services_list(ConnmanManager *self, GHashTable *list)
{
  g_hash_table_foreach(list, notify_service_removal, self);
  g_hash_table_destroy(list);
}

static void update_services(ConnmanManager *self, GVariant *variant)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  GHashTable *old_services;
  ConnmanService *service;
  gchar *path;
  GVariantIter iter;
  gint count;

  old_services = priv->services;
  priv->services = create_services_list(self);

  g_variant_iter_init(&iter, variant);
  count = 0;

  while (g_variant_iter_next(&iter, "o", &path)) {
    /* try find an existing service */
    if (old_services != NULL) {
      service = g_hash_table_lookup(old_services, path);

      if (service != NULL) {
	/* there's already an object for this service */
	g_hash_table_remove(old_services, path);
	g_hash_table_insert(priv->services, path, service);
	goto next;
      }
    }

    service = create_service(self, path);
    g_free(path);

  next:
    if (count++ == 0) {
      if (priv->pending_default_service != NULL)
	g_object_unref(priv->pending_default_service);

      priv->pending_default_service = g_object_ref(service);
      update_default_service(self);
    }
  }

  /* remove services which are not visible anymore */
  if (old_services != NULL) {
    remove_services_list(self, old_services);
    old_services = NULL;
  }
}

static void update_enabled_technologies(ConnmanManager *self,
					GVariant *variant)
{
}

static void update_property(ConnmanManager *self, const gchar *property,
			    GVariant *variant)
{
  if (g_strcmp0(property, CONNMAN_PROPERTY_SERVICES) == 0) {
    update_services(self, variant);
  }
  else if (g_strcmp0(property, CONNMAN_PROPERTY_ENABLED_TECHNOLOGIES) == 0) {
    update_enabled_technologies(self, variant);
  } else {
    /* unknown property */
    g_variant_unref(variant);
    variant = NULL;
  }
}

static void property_changed(ConnmanManager *self, GVariant *parameters)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  GVariant *variant;
  gchar *name;

  g_return_if_fail(priv != NULL);

  g_variant_get(parameters, "(sv)", &name, &variant);

  /* callee takes the ownership of the variant */
  update_property(self, name, variant);

  g_free(name);
}

static void connman_manager_g_signal(GDBusProxy *proxy,
				   const gchar *sender_name,
				   const gchar *signal_name,
				   GVariant *parameters,
				   gpointer user_data)
{
  ConnmanManager *self = CONNMAN_MANAGER(user_data);

  /*
   * gdbus documentation is not clear who owns the variant so take it, just
   * in case
   */
  g_variant_ref(parameters);

  if (g_strcmp0(signal_name, "PropertyChanged") == 0)
    property_changed(self, parameters);

  g_variant_unref(parameters);
}

static void get_properties_cb(GObject *source_object, GAsyncResult *res,
			      gpointer user_data)
{
  ConnmanManager *self = CONNMAN_MANAGER(user_data);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  GVariant *result, *variant, *dict;
  GError *error = NULL;
  GVariantIter iter, dict_iter;
  gchar *name;

  g_return_if_fail(self != NULL);
  g_return_if_fail(priv != NULL);

  result = g_dbus_proxy_call_finish(priv->proxy, res, &error);

  if (error != NULL) {
    g_warning("Failed to get connman properties: %s", error->message);
    g_error_free(error);
    return;
  }

  if (result == NULL)
    return;

  g_variant_iter_init(&iter, result);

  /* get the dict */
  dict = g_variant_iter_next_value(&iter);

  /* go through the dict */
  g_variant_iter_init(&dict_iter, dict);

  while (g_variant_iter_next(&dict_iter, "{sv}", &name, &variant)) {
    /* callee owns the variant now */
    update_property(self, name, variant);

    g_free(name);
  }

  if (!priv->initialised)
    priv->initialised = TRUE;

  if (!priv->connected && g_hash_table_size(priv->services) == 0) {
    /* no services so we can change to connected immediately */
    priv->connected = TRUE;
    g_object_notify(G_OBJECT(self), "connected");
  }

  g_variant_unref(dict);
  g_variant_unref(result);
}


static void get_properties(ConnmanManager *self)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  g_return_if_fail(self != NULL);

  g_dbus_proxy_call(priv->proxy, "GetProperties", NULL,
  		    G_DBUS_CALL_FLAGS_NONE, -1, NULL,
  		    get_properties_cb, self);
}

static void connman_appeared(GDBusConnection *connection, const gchar *name,
			     const gchar *name_owner, gpointer user_data)
{
  ConnmanManager *self = CONNMAN_MANAGER(user_data);

  get_properties(self);
}


static void connman_vanished(GDBusConnection *connection, const gchar *name,
			     gpointer user_data)
{
  ConnmanManager *self = CONNMAN_MANAGER(user_data);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  /* if connman is not available, send a disconnected signal during init */

  if (priv->initialised && !priv->connected)
    return;

  priv->connected = FALSE;

  if (priv->services != NULL) {
    remove_services_list(self, priv->services);
    priv->services = NULL;
  }

  if (!priv->initialised)
    priv->initialised = TRUE;

  g_object_notify(G_OBJECT(self), "connected");
}


static void create_proxy_cb(GObject *source_object, GAsyncResult *res,
			    gpointer user_data)
{
  ConnmanManager *self = CONNMAN_MANAGER(user_data);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);
  GDBusNodeInfo *introspection_data;
  GDBusInterfaceInfo *info;
  GError *error = NULL;

  g_return_if_fail(priv != NULL);

  priv->proxy = g_dbus_proxy_new_finish(res, &error);

  if (error != NULL) {
    g_warning("Failed to get connman proxy: %s", error->message);
    g_error_free(error);
    return;
  }

  if (priv->proxy == NULL) {
    g_warning("Failed to get connman proxy, but no errors");
    return;
  }

  g_signal_connect(priv->proxy, "g-signal", G_CALLBACK(connman_manager_g_signal),
		   self);

  introspection_data = g_dbus_node_info_new_for_xml(connman_manager_xml,
						    NULL);
  g_return_if_fail(introspection_data != NULL);

  info = introspection_data->interfaces[0];
  g_dbus_proxy_set_interface_info(priv->proxy, info);

  priv->watch_id = g_bus_watch_name(G_BUS_TYPE_SYSTEM, CONNMAN_SERVICE_NAME,
				    G_BUS_NAME_WATCHER_FLAGS_NONE,
				    connman_appeared,
				    connman_vanished,
				    self,
				    NULL);
}

static void set_property(GObject *object, guint property_id,
			 const GValue *value, GParamSpec *pspec)
{
  ConnmanManager *self = CONNMAN_MANAGER(object);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  g_return_if_fail(priv != NULL);

  switch(property_id) {
  case PROP_CONNECTED:
      priv->connected = g_value_get_boolean(value);
      break;
  case PROP_DEFAULT_SERVICE:
      priv->default_service = g_value_get_object(value);
      break;
  default:
    /* We don't have any other property... */
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
    break;
  }
}

static void get_property(GObject *object, guint property_id,
			 GValue *value, GParamSpec *pspec)
{
  ConnmanManager *self = CONNMAN_MANAGER(object);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  switch(property_id) {
  case PROP_CONNECTED:
    g_value_set_boolean(value, priv->connected);
    break;
  case PROP_DEFAULT_SERVICE:
    g_value_set_object(value, priv->default_service);
    break;
  default:
    /* We don't have any other property... */
    G_OBJECT_WARN_INVALID_PROPERTY_ID(object, property_id, pspec);
    break;
  }
}

static void connman_manager_dispose(GObject *object)
{
  ConnmanManager *self = CONNMAN_MANAGER(object);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  if (priv->services != NULL) {
    remove_services_list(self, priv->services);
    priv->services = NULL;
  }

  if (priv->proxy != NULL) {
    g_object_unref(priv->proxy);
    priv->proxy = NULL;
  }

  if (priv->watch_id != 0) {
    g_bus_unwatch_name(priv->watch_id);
    priv->watch_id = 0;
  }

  if (priv->default_service != NULL) {
    g_object_unref(priv->default_service);
    priv->default_service = NULL;
  }

  if (priv->pending_default_service != NULL) {
    g_object_unref(priv->pending_default_service);
    priv->pending_default_service = NULL;
  }

  G_OBJECT_CLASS(connman_manager_parent_class)->dispose(object);
}

static void connman_manager_finalize(GObject *object)
{
  ConnmanManager *self = CONNMAN_MANAGER(object);
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  g_free(priv->array);
  priv->array = NULL;

  G_OBJECT_CLASS(connman_manager_parent_class)->finalize(object);
}

static void connman_manager_class_init(ConnmanManagerClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
  GParamSpec *pspec;
  guint id;

  g_type_class_add_private(klass, sizeof(ConnmanManagerPrivate));

  gobject_class->dispose = connman_manager_dispose;
  gobject_class->finalize = connman_manager_finalize;
  gobject_class->set_property = set_property;
  gobject_class->get_property = get_property;

  id = g_signal_new("service-added",
		    G_TYPE_FROM_CLASS(klass),
		    G_SIGNAL_RUN_LAST,
		    0, NULL, NULL,
		    _marshal_VOID__BOXED,
		    G_TYPE_NONE, 1, CONNMAN_TYPE_SERVICE);
  signals[SERVICE_ADDED_SIGNAL] = id;

  id = g_signal_new("service-removed",
		    G_TYPE_FROM_CLASS(klass),
		    G_SIGNAL_RUN_LAST,
		    0, NULL, NULL,
		    _marshal_VOID__STRING,
		    G_TYPE_NONE, 1, G_TYPE_STRING);
  signals[SERVICE_REMOVED_SIGNAL] = id;

  pspec = g_param_spec_boolean("connected",
			       "ConnmanManager's connected property",
			       "Set connected state", FALSE,
			       G_PARAM_READABLE);
  g_object_class_install_property(gobject_class, PROP_CONNECTED, pspec);

  pspec = g_param_spec_object("default-service",
			     "ConnmanManager's default service",
			     "Set default service",
			     CONNMAN_TYPE_SERVICE,
			     G_PARAM_READABLE);
  g_object_class_install_property(gobject_class, PROP_DEFAULT_SERVICE, pspec);
}

static void connman_manager_init(ConnmanManager *self)
{
  ConnmanManagerPrivate *priv = GET_PRIVATE(self);

  priv->services = NULL;
  priv->proxy = NULL;
  priv->connected = FALSE;
  priv->initialised = FALSE;

  priv->default_service = NULL;
  priv->pending_default_service = NULL;

  g_dbus_proxy_new_for_bus(G_BUS_TYPE_SYSTEM, G_DBUS_PROXY_FLAGS_NONE,
			   NULL, CONNMAN_SERVICE_NAME, CONNMAN_MANAGER_PATH,
			   CONNMAN_MANAGER_INTERFACE, NULL,
			   create_proxy_cb, self);
}

ConnmanManager *connman_manager_new(void)
{
  return g_object_new(CONNMAN_TYPE_MANAGER, NULL);
}
