/*
 * indicator-network - user interface for connman
 * Copyright 2010-2011 Canonical Ltd.
 *
 * Authors:
 * Andrew Higginson
 * Gordon Allott <gord.allott@canonical.com>
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

namespace Utils.Color {
    /* from libunity */
    public static void rgb_to_hsv (float r, float g, float b,
                                   out float hue, out float sat, out float val)
    {
        float min, max = 0.0f;
        if (r > g)
            max = (r > b) ? r : b;
        else
            max = (g > b) ? g : b;
        if (r < g)
            min = (r < b) ? r : b;
        else
            min = (g < b) ? g : b;

        val = max;

        float delta = max - min;
        if (delta > 0.000001)
        {
            sat = delta / max;
            hue = 0.0f;
            if (r == max)
            {
                hue = (g - b) / delta;
                if (hue < 0.0f)
                    hue += 6.0f;
            }
            else if (g == max)
            {
                hue = 2.0f + (b - r) / delta;
            }
            else if (b == max)
            {
                hue = 4.0f + (r - g) / delta;
            }
            hue /= 6.0f;
        }
        else
        {
            sat = 0.0f;
            hue = 0.0f;
        }
    }

    /* from libunity */
    public static void hsv_to_rgb (float hue, float sat, float val,
                                   out float r, out float g, out float b)
    {
        int    i;
        float f, w, q, t;

        if (sat == 0.0)
        {
            r = g = b = val;
        }
        else
        {
            if (hue == 1.0)
                hue = 0.0f;

            hue *= 6.0f;

            i = (int) hue;
            f = hue - i;
            w = val * (1.0f - sat);
            q = val * (1.0f - (sat * f));
            t = val * (1.0f - (sat * (1.0f - f)));

            switch (i)
            {
            case 0:
                r = val;
                g = t;
                b = w;
                break;
            case 1:
                r = q;
                g = val;
                b = w;
                break;
            case 2:
                r = w;
                g = val;
                b = t;
                break;
            case 3:
                r = w;
                g = q;
                b = val;
                break;
            case 4:
                r = t;
                g = w;
                b = val;
                break;
            case 5:
                r = val;
                g = w;
                b = q;
                break;
            }
        }
    }

    static void color_to_hls(Gdk.Color color, out float hue, out float luminance, out float saturation) {
        float red, green, blue;
        float h, l, v;
        
        red = color.red / 65535.0f;
        green = color.green / 65535.0f;
        blue = color.blue / 65535.0f;

        rgb_to_hsv(red, green, blue, out h, out l, out v);

        hue = h;
        luminance = l;
        saturation = v;
    }

    static void color_from_hls(float hue, float luminance, float saturation, out Gdk.Color color) {
        float red, green, blue;

        hsv_to_rgb(hue, luminance, saturation, out red, out green, out blue);

        color.red = (uint16) Math.floorf(red * 65535.0f + 0.5f);
        color.green = (uint16) Math.floorf(green * 65535.0f + 0.5f);
        color.blue = (uint16) Math.floorf(blue * 65535.0f + 0.5f);
    }

    static Gdk.Color shade(Gdk.Color color) {
        float h;
        float l;
        float s;
        float factor;
        Utils.Color.color_to_hls(color, out h, out l, out s);

        if (l > 0.5) {
            // Colour is light, lets darken it
            factor = 0.6f;
        } else {
            // Colour is dark, lets lighten it
            factor = 1.4f;
        }
    
        l *= factor;
        if (l > 1.0f) {
            l = 1.0f;
        } else if (l < 0) {
            l = 0;
        }
    
        s*= factor;
        if (s > 1.0f) {
            s= 1.0f;
        } else if (s < 0) {
            s = 0;
        }

        Gdk.Color new_color;
        Utils.Color.color_from_hls(h, l, s, out new_color);
        
        return new_color;
    }
}

namespace Utils.Gui {
    static Gtk.Builder new_builder(string path) {
        var builder = new Gtk.Builder();
        try {
            builder.add_from_file(path);
        } catch (Error e) {
            logger.debug(@"Can't find $path.");
        }
        builder.set_translation_domain("indicator-network-settings");
        return builder;
    }
}
