/*
 * indicator-network - user interface for connman
 * Copyright 2010-2011 Canonical Ltd.
 *
 * Authors:
 * Andrew Higginson
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

class MobileBox : DeviceBox {

    /* A Gtk.VBox which holds all of the widgets to control a mobile
      device. It contains an InfoBox and ToggleSwitch to control and
      monitor the state of the mobile device, and a set of widgets
      to enter details about the mobile network to connect to. */

    private Device device;
    private InfoBox infobox;
    private ToggleSwitch toggleswitch;
    private Gtk.Label label_status;
    private Gtk.Builder builder;
    private Gtk.Table table_settings;
    private Gtk.Entry entry_pin;
    private Gtk.CheckButton checkbutton_show_pin;

    private bool device_editable;
    private bool settings_editable;

    private const string[] sensitive_widgets = {"label_apn_place", "label_apn",
                                        "label_pin_auth", "checkbutton_show_pin"};

    public MobileBox(Device device, string datadir) {
        this.set_spacing(12);
        this.device = device;
        this.device.notify["state"].connect(this.on_device_state_changed);

        // Infobox and Togglswitch
        /// Creation
        this.infobox = new InfoBox(false, 12);
        this.toggleswitch = new ToggleSwitch();
        this.label_status = new Gtk.Label(null);
        /// Padding and alignment
        this.label_status.set_alignment(0, 0.5f);
        this.infobox.set_border_width(10);
        /// Packing
        this.infobox.pack_start(this.label_status, true, true);
        this.infobox.pack_start(this.toggleswitch, false, true);
        this.pack_start(this.infobox, false, false);
        /// Connect signals;
        this.toggleswitch.toggled.connect(this.on_toggleswitch_toggled);

        // Network Settings
        /// Creation
        this.builder = Utils.Gui.new_builder(datadir+"ui/mobile_box.ui");
        this.get_widgets();
        this.connect_signals();
        /// Packing
        this.pack_start(this.table_settings, true, true);

        this.update_widget_states(this.device.state);
        this.on_checkbutton_show_pin_toggled();
    }

    private void get_widgets() {
        this.table_settings = this.builder.get_object("table_settings") as Gtk.Table;
        this.entry_pin = this.builder.get_object("entry_pin") as Gtk.Entry;
        this.checkbutton_show_pin = this.builder.get_object("checkbutton_show_pin") as Gtk.CheckButton;
    }
    
    private void connect_signals() {
        this.checkbutton_show_pin.toggled.connect(this.on_checkbutton_show_pin_toggled);
    }

    private void update_widget_states(DeviceState state) {
        bool toggleswitch_state = false;
        string status_text = "";

        switch (state) {
            case DeviceState.CONNECTED:
                this.device_editable = true;
                this.settings_editable = false;
                toggleswitch_state = true;
                status_text = ("Connected to “Vodafone UK”.");
                break;
            case DeviceState.ONLINE:
                this.device_editable = true;
                this.settings_editable = false;
                toggleswitch_state = true;
                status_text = ("Connected to the Internet on “Vodafone UK”.");
                break;
            case DeviceState.OFF:
                this.device_editable = false;
                this.settings_editable = true;
                toggleswitch_state = false;
                status_text = ("The Mobile Broadband device is powered off.");
                break;
            case DeviceState.OFFLINE:
                this.device_editable = true;
                this.settings_editable = true;
                toggleswitch_state = false;
                status_text = ("Mobile Broadband is offline.");
                break;
        }
            
        this.toggleswitch.set_sensitive(this.device_editable);
        this.toggleswitch.set_active(toggleswitch_state);
        this.label_status.set_text(status_text);

        this.table_settings.foreach(this.sensitise_widgets);
        }

    private void sensitise_widgets(Gtk.Widget widget) {
        
        var name = widget.get_name();
        if (name in this.sensitive_widgets) {
            widget.set_sensitive(this.device_editable);
        } else if (!(widget is Gtk.Box)) {
            widget.set_sensitive(this.settings_editable);
        } else {
            var container = widget as Gtk.Container;
            container.foreach(this.sensitise_widgets);
        }
    }
    
    public override void on_device_state_changed(ParamSpec p) {
        this.update_widget_states(this.device.state);
    }
    
    private void on_toggleswitch_toggled() {
        if (this.toggleswitch.get_active()) {
            this.device.state = DeviceState.CONNECTED;
        } else {
            this.device.state = DeviceState.OFFLINE;
        }
    }

    private void on_checkbutton_show_pin_toggled() {
        this.entry_pin.set_visibility(this.checkbutton_show_pin.get_active());
    }
}
